#' Angle based clustering
#'
#' @details
#' Given the adjacency spectral embeding X, perform
#' the angle-based clustering procedure (Algorithm 2) in Lyzinski et al.
#'
#' @param X an input matrix
#' @param R a sample size
#' @param seed a random seed
#' @param verbose a logical flag to allow to show clustering information on progress
#' @param clusinfo logical 
#' 
#' @return a list containing of elements:
#' \item{memb}{a clustering membership vector.}
#' \item{asw}{an averaged silhouette width.}
#' \item{clusinfo}{various clustering statistics including size,
#' maximum dissimilarity, averaged dissimilarity, diameter, separation among the clusters.}
#' @author Youngser Park
#' @references Vince Lyzinski, Minh Tang, Avanti Athreya, Youngser Park, Carey E. Priebe,
#' "Community detection and classification for hierarchical stochastic blockmodels,"
#' \emph{IEEE Transactions on Network Science and Engineering}, accepted for publication, 2016.
#' @export

vlclust <- function(X,R=2,seed=1,verbose=FALSE,clusinfo=FALSE)
{
    (n <- nrow(X))

    if (length(R)>1) stop("Please use vclustK instead!")
    if (n < R) stop("R should be <= n!")

    set.seed(seed)
    samp <- sample(n,R)
    S <- X[samp,]
    yz <- S %*% t(S)
    diag(yz) <- yz[lower.tri(yz)] <- 0
    maxyz <- max(yz)
    argmax <- which(yz==maxyz,arr.ind=TRUE)
    y <- argmax[1]

    ## initialize cluster
    if (verbose) cat("  R =", R, ", seed =", seed, "  init...")
    for (i in 1:n) {
        if (verbose & n>10000000 & (i %% 10000000==0)) {
            cat(i/1000000, "M...")
        }
        Xis <- X[i,] %*% t(S)
        if (max(Xis) < maxyz) {
            S <- S[-y,] # remove y from S
            S <- rbind(S,X[i,]) # add Xi to S
            yz <- S %*% t(S)
            diag(yz) <- yz[lower.tri(yz)] <- 0
            maxyz <- max(yz)
            argmax <- which(yz==maxyz,arr.ind=TRUE)
            y <- argmax[1]
        }
    }
    if (verbose & n>1000000 & (i %% 1000000==0)) {
        cat("\n")
    }

    ## do assign Xi into C
    memb <- sapply(1:n, function(x) which.max(X[x,] %*% t(S)))
    cl <- sort(unique(memb))

    ## calc Silhouette
    ## (b-a) / max(b,a)
    ## a: avg dist from i to everyone within cluster
    ## b: min(avg dist from i to closest cluster)
##     si <- rep(0,n)
##     for (i in 1:n) {
##         ai <- mean(-(X[i,] %*% t(X[memb==memb[i],])))
##         bi <- min(sapply(cl[-memb[i]], function(x) mean(-(X[i,] %*% t(X[memb==x,])))))
##         si[i] <- (bi-ai) / abs(min(ai,bi))
## #        ai <- mean(rect.dist(X[i,,drop=FALSE], X[memb==memb[i],]))
## #        bi <- min(sapply(cl[-memb[i]], function(x) mean(rect.dist(X[i,,drop=FALSE],X[memb==x,]))))
## #        si[i] <- (bi-ai) / max(ai,bi)
## #        if (verbose) cat("(i,a,b,s) = (",i,",",ai,",",bi,",",si[i],"\n")
##     }
##     asw <- mean(si)

    ## new asw
    tmp <- (S %*% t(S)) / (outer(sqrt(rowSums(S^2)),sqrt(rowSums(S^2)), FUN="pmin"))
    diag(tmp) <- -1e5
    (asw <- max(tmp))
    if (verbose) cat("seed = ", seed, ", R = ", R, ",   asw = ", asw, ", C =", cl,"\n")

# plot(X,pch=".",col=memb+1)
# points(S,col=1,pch=21,bg=2:19)


    ## clusinfo
    if (clusinfo) {
        df <- data.frame(X=X,C=memb)
        split.df <- split(df[,-ncol(df)],df[,ncol(df)])
        (size <- sapply(split.df, nrow))
        cent <- t(sapply(split.df, colMeans))
        rownames(cent) <- cl
#        cl <- 1:R
        (maxdiss <- sapply(1:length(cl), function(x) max(rect.dist.sqrt(cent[x,,drop=FALSE],split.df[[x]]))))
        (avgdiss <- sapply(1:length(cl), function(x) mean(rect.dist.sqrt(cent[x,,drop=FALSE],split.df[[x]]))))
        (diameter <- sapply(split.df, function(x) max(dist(x))))
#        (separation <- sapply(cl, function(x) sapply(cl[-x], function(y) min(rect.dist.sqrt(split.df[[x]], split.df[[y]])))))
        cl2 <- 1:length(cl)
        (separation <- sapply(cl2, function(x) min(sapply(cl2[-x], function(y) min(rect.dist.sqrt(split.df[[x]], split.df[[y]]))))))

        clusinfo <- data.frame(size=size,maxdiss=maxdiss,avgdiss=avgdiss,diameter=diameter,separation=separation)
    } else {
        clusinfo <- NULL
    }

    return(list(memb=memb,asw=asw,clusinfo=clusinfo))
}

vlclustK <- function(X,krange=2,seed=1,verbose=FALSE,clusinfo=FALSE,aswonly=FALSE)
{
    (n <- nrow(X))
    (numR <- length(krange))
    if (!aswonly) Cmat <- matrix(0,n,numR)
    svec <- rep(0,numR)
    clinfo <- NULL
    ind <- 1
    for (R in krange) {
        out <- vlclust(X,R=R,seed=seed,verbose=verbose,clusinfo=clusinfo)
        if (!aswonly) Cmat[,ind] <- out$memb
        svec[ind] <- out$asw
        clinfo[[ind]] <- out$clusinfo
        ind <- ind+1
    }
    svec[svec<0] <- 0

    ## find Rhat
    if (aswonly) {
        memb <- clusinfo <- NULL
    } else {
        df <- data.frame(R=krange,x=rep(1,length(svec)),asw=svec)
        df <- do.call(rbind, by(df, df$x, function(y) {
            y$ratio <- exp(c(tail(y$asw, -1),NA)) / exp(y$asw)
            y
        }))
        Rargmax <- ifelse(numR>1,which.max(df$ratio),1)
        memb <- Cmat[,Rargmax]
        clusinfo <- clinfo[[Rargmax]]
    }

    return(list(memb=memb,asw=svec,clusinfo=clusinfo))
}

vlclustpar <- function(X,krange=2,seed=1,ncore=7,verbose=FALSE,clusinfo=FALSE,aswonly=FALSE)
{
    suppressMessages(require(parallel))
    suppressMessages(require(plyr))

    out <- mclapply(krange, function(x) vlclust(X,x,seed,verbose=verbose,clusinfo=clusinfo), mc.cores=ncore)
#    save(out,file=paste0("vlout-Rmax",max(krange),"-seed",seed,".Rbin"))
    if (!aswonly) Cmat <- as.matrix(ldply(sapply(out,"[",1))[,-1])
    svec <- unlist(sapply(out,"[",2)); names(svec) <- krange
    svec[svec<0] <- 0

    ## find Rhat
    if (aswonly) {
        memb <- clusinfo <- NULL
    } else {
        clinfo <- sapply(out,"[",3)
        df <- data.frame(R=krange,x=rep(1,length(svec)),asw=svec)
        df <- do.call(rbind, by(df, df$x, function(y) {
            y$ratio <- exp(c(tail(y$asw, -1),NA)) / exp(y$asw)
            y
        }))
        Rargmax <- which.max(df$ratio)
        memb <- as.numeric(Cmat[Rargmax,])
        clusinfo <- clinfo[[Rargmax]]
    }


    return(list(memb=memb,asw=svec,clusinfo=clusinfo))

}

